import { columnTypes } from "./encoding.js";
import { Method, Query, quote } from "./utils.js";
export class RelationAPI {
    db;
    config;
    relation;
    table;
    sourceIndex;
    targetIndex;
    sourceColumnNames;
    targetColumnNames;
    #select;
    #insert;
    #delete;
    #clear;
    constructor(db, config, relation, clear) {
        this.db = db;
        this.config = config;
        this.relation = relation;
        this.table = `${relation.source}/${relation.sourceProperty}`;
        this.sourceIndex = `${relation.source}/${relation.sourceProperty}/source`;
        this.targetIndex = `${relation.source}/${relation.sourceProperty}/target`;
        this.sourceColumnNames = [];
        this.targetColumnNames = [];
        const sourcePrimaryKey = config.primaryKeys[relation.source];
        const targetPrimaryKey = config.primaryKeys[relation.target];
        {
            const columns = [];
            if (sourcePrimaryKey.length === 1) {
                const [{ type }] = sourcePrimaryKey;
                columns.push(`_source ${columnTypes[type]} NOT NULL`);
                this.sourceColumnNames.push("_source");
            }
            else {
                for (const [i, { type }] of sourcePrimaryKey.entries()) {
                    const columnName = `_source/${i}`;
                    columns.push(`"${columnName}" ${columnTypes[type]} NOT NULL`);
                    this.sourceColumnNames.push(columnName);
                }
            }
            if (targetPrimaryKey.length === 1) {
                const [{ type }] = targetPrimaryKey;
                columns.push(`_target ${columnTypes[type]} NOT NULL`);
                this.targetColumnNames.push("_target");
            }
            else {
                for (const [i, { type }] of targetPrimaryKey.entries()) {
                    const columnName = `_target/${i}`;
                    columns.push(`"${columnName}" ${columnTypes[type]} NOT NULL`);
                    this.targetColumnNames.push(columnName);
                }
            }
            if (clear) {
                db.exec(`DROP TABLE IF EXISTS "${this.table}"`);
            }
            db.exec(`CREATE TABLE IF NOT EXISTS "${this.table}" (${columns.join(", ")})`);
        }
        const sourceColumns = this.sourceColumnNames.map(quote).join(", ");
        db.exec(`CREATE INDEX IF NOT EXISTS "${this.sourceIndex}" ON "${this.table}" (${sourceColumns})`);
        if (relation.indexed) {
            this.addTargetIndex();
        }
        // Prepare methods
        const insertColumns = [...this.sourceColumnNames, ...this.targetColumnNames].map(quote).join(", ");
        const insertParamCount = this.sourceColumnNames.length + this.targetColumnNames.length;
        const insertParams = Array.from({ length: insertParamCount }).fill("?").join(", ");
        this.#insert = new Method(this.db, `INSERT INTO "${this.table}" (${insertColumns}) VALUES (${insertParams})`);
        const selectBySource = this.sourceColumnNames.map((name) => `"${name}" = ?`).join(" AND ");
        this.#delete = new Method(this.db, `DELETE FROM "${this.table}" WHERE ${selectBySource}`);
        this.#clear = new Method(this.db, `DELETE FROM "${this.table}"`);
        // Prepare queries
        const targetColumns = this.targetColumnNames.map(quote).join(", ");
        this.#select = new Query(this.db, `SELECT ${targetColumns} FROM "${this.table}" WHERE ${selectBySource}`);
    }
    drop() {
        this.db.exec(`DROP TABLE "${this.table}"`);
    }
    addTargetIndex() {
        const targetColumns = this.targetColumnNames.map(quote).join(", ");
        this.db.exec(`CREATE INDEX IF NOT EXISTS "${this.targetIndex}" ON "${this.table}" (${targetColumns})`);
    }
    removeTargetIndex() {
        this.db.exec(`DROP INDEX IF EXISTS "${this.targetIndex}"`);
    }
    get(sourceKey) {
        const targets = this.#select.all(sourceKey);
        return targets.map((record) => this.targetColumnNames.map((name) => record[name]));
    }
    add(sourceKey, targetKeys) {
        for (const targetKey of targetKeys) {
            this.#insert.run([...sourceKey, ...targetKey]);
        }
    }
    delete(sourceKey) {
        this.#delete.run(sourceKey);
    }
    clear() {
        this.#clear.run([]);
    }
}
